import datetime
import traceback

from django.utils.timezone import now, localtime
from django import template

from wcs.qommon.misc import unlazy
from wcs.sql import get_connection, atomic

register = template.Library()


def error(result, err_desc):
    result.update({'err': 1, 'err_desc': err_desc})
    return result


def ok(result, slot, **kwargs):
    result.pop('err_desc')
    result.update({'err': 0, 'slot': slot})
    result.update(kwargs)
    return result


def main(result, identifier, name, count, per_minutes):
    ref = localtime(now()).replace(hour=0, minute=0, second=0, microsecond=0)

    conn = get_connection(new=True)
    with conn.cursor() as cur:
        # create table
        cur.execute(
            '''\
CREATE TABLE IF NOT EXISTS waiting_queue (
    name TEXT,
    userid TEXT,
    created TIMESTAMPTZ,
    slot TIMESTAMPTZ)'''
        )

        # check if date_bin exists or install it
        cur.execute('SELECT EXISTS(SELECT * FROM pg_proc WHERE proname = %s)', ['waiting_queue_date_bin'])
        if not cur.fetchone()[0]:
            cur.execute(
                '''CREATE OR REPLACE FUNCTION waiting_queue_date_bin
 (trunc_period INTERVAL, ts TIMESTAMPTZ, base_ts TIMESTAMPTZ DEFAULT 'epoch')
 RETURNS TIMESTAMPTZ LANGUAGE SQL IMMUTABLE AS
$$
SELECT
  base_ts
  + FLOOR(EXTRACT(epoch FROM ts - base_ts) / EXTRACT(epoch FROM trunc_period))::bigint
  * trunc_period;
$$;'''
            )

        with atomic():
            # flush old slots
            cur.execute('DELETE FROM waiting_queue WHERE created < %s OR slot < %s', [ref, ref])
            # lock the table during computations
            cur.execute('LOCK waiting_queue')
            # find slot
            cur.execute(
                'SELECT slot, created FROM waiting_queue WHERE name = %s AND userid = %s', [name, identifier]
            )
            row = cur.fetchone()
            if row:
                slot = localtime(row[0])
                created = localtime(row[1])
                cur.execute('SELECT count(*) FROM waiting_queue WHERE name = %s AND slot = %s', [name, slot])
                count = cur.fetchone()[0]
                return ok(result, slot, case='existing', current_count=count, created=created, ref=ref)

            # allocate new slot
            # 1. get current bin
            query = '''\
SELECT
    waiting_queue_date_bin(
        (%s::text || ' minutes')::interval,
        GREATEST(MAX(slot), NOW()),
        TIMESTAMP %s)
FROM
    waiting_queue
WHERE
    name = %s
    '''
            cur.execute(query, [per_minutes, ref, name])
            current_bin = localtime(cur.fetchone()[0])
            result['current_bin'] = current_bin
            # 2. get number of user in bin
            query = '''
SELECT
    COUNT(*)
FROM
    waiting_queue
WHERE
    waiting_queue_date_bin(
        (%s::text || ' minutes')::interval,
        slot,
        TIMESTAMP %s) = %s
    AND name = %s'''
            cur.execute(query, [per_minutes, ref, current_bin, name])
            current_count = cur.fetchone()[0]
            result['current_count'] = current_count
            # 3. compute new slot
            slot = current_bin
            if current_count >= count:
                slot += datetime.timedelta(minutes=per_minutes)
            # 3. insert the slot
            cur.execute(
                '''
INSERT INTO
    waiting_queue(
        name,
        userid,
        created,
        slot)
VALUES (
    %s,
    %s,
    NOW(),
    %s) RETURNING created''',
                [name, identifier, slot],
            )
            created = localtime(cur.fetchone()[0])
    return ok(result, slot, case='allocated', created=created, ref=ref)


@register.filter
def waiting_queue(identifier, arg):
    result = {
        'err': 1,
        'err_desc': 'Script does not run',
    }

    if not identifier:
        return error(
            result,
            'First argument should identify user and cannot be empty.'
            + ' Example: the "session_user_nameid" variable may be used.',
        )

    try:
        name, count, per_minutes = arg.split('/')
        count = int(count)
        per_minutes = int(per_minutes)
    except ValueError:
        return error(
            result,
            'Invalid argument: "%s".' % unlazy(arg)
            + " Second argument must contain, separated by slashes:"
            + ' the name of the queue,'
            + ' the number of allocable slots on the queue,'
            + ' and the time in minutes after wich new slots will be allocated.'
            + ' Example: "my_queue/100/10" gives access to 100 new users every 10 minutes',
        )

    result.update(
        {
            'identifier': identifier,
            'arg': unlazy(arg),
            'name': name,
            'count': count,
            'per_minutes': per_minutes,
        }
    )

    try:
        return main(result, identifier, name, count, per_minutes)
    except Exception:
        tb = traceback.format_exc()
        return error(result, f'Exception: {tb}')
